@@ -1,42 +1,40 @@
-import attrs
+import logging
 import typing as ty
 from collections import OrderedDict
-import logging
 from copy import copy
-from typing_extensions import Self
+
+import attrs
 import attrs.converters
-from fileformats.core import DataType
+from fileformats.core import DataType, FileSet, to_mime
 from fileformats.core.exceptions import FormatConversionError
-from pydra.utils.typing import StateArray
-from pydra.utils import get_fields
+from pydra.compose import python, workflow
 from pydra.compose.base import Task
-from pydra.compose import workflow, python
-from frametree.core.exceptions import (
-    FrameTreeNameError,
-    FrameTreeUsageError,
-    FrameTreeDesignError,
-    FrameTreePipelinesStackError,
-    FrameTreeOutputNotProducedException,
-    FrameTreeDataMatchError,
-)
+from pydra.utils import get_fields
+from pydra.utils.typing import StateArray, TypeParser, is_union
+from typing_extensions import Self
+
 import frametree.core.frameset.base
 import frametree.core.row
 from frametree.core.axes import Axes
-from frametree.core.utils import (
-    path2varname,
-    add_exc_note,
+from frametree.core.column import SinkColumn
+from frametree.core.exceptions import (
+    FrameTreeDataMatchError,
+    FrameTreeDesignError,
+    FrameTreeNameError,
+    FrameTreeOutputNotProducedException,
+    FrameTreePipelinesStackError,
+    FrameTreeUsageError,
 )
+from frametree.core.frameset.base import FrameSet
 from frametree.core.serialize import (
+    ClassResolver,
+    ObjectListConverter,
     asdict,
     fromdict,
     pydra_asdict,
     pydra_fromdict,
-    ClassResolver,
-    ObjectListConverter,
 )
-from frametree.core.column import SinkColumn
-from frametree.core.frameset.base import FrameSet
-
+from frametree.core.utils import add_exc_note, path2varname
 
 logger = logging.getLogger("frametree")
 
@@ -62,6 +60,7 @@ class PipelineField:
         converter=ClassResolver(
             DataType,
             allow_none=True,
+            allow_optional=True,
             alternative_types=[frametree.core.row.DataRow],
         ),
     )
@@ -173,8 +172,8 @@ class Pipeline:
                 # Check that a converter can be found if required
                 if (
                     outpt.datatype
-                    and not issubclass(outpt.datatype, column.datatype)
-                    and not issubclass(column.datatype, outpt.datatype)
+                    and not TypeParser.is_subclass(outpt.datatype, column.datatype)
+                    and not TypeParser.is_subclass(column.datatype, outpt.datatype)
                 ):
                     try:
                         column.datatype.get_converter(outpt.datatype)
@@ -421,7 +420,7 @@ def PipelineRowWorkflow(
     inputs: ty.List[PipelineField],
     outputs: ty.List[PipelineField],
     converter_args: ty.Dict[str, dict],
-):
+) -> str:
 
     # Get the values from the frameset, caching remote files locally
     source = workflow.add(
@@ -477,37 +476,57 @@ def PipelineRowWorkflow(
 
     # Do input datatype conversions if required
     for inpt in inputs:
-        if inpt.datatype == frametree.core.row.DataRow:
+        if (
+            inpt.datatype == frametree.core.row.DataRow
+            or not inpt.datatype
+            or is_coercible(inpt.datatype, frameset[inpt.name].datatype)
+        ):
+            # No conversion required
             continue
         stored_format = frameset[inpt.name].datatype
-        if (
-            inpt.datatype
-            and not issubclass(inpt.datatype, stored_format)
-            and not issubclass(stored_format, inpt.datatype)
-        ):
-            converter = inpt.datatype.get_converter(stored_format)
-            logger.info(
-                "Adding implicit conversion for input '%s' from %s to %s",
-                inpt.name,
-                stored_format.mime_like,
-                inpt.datatype.mime_like,
+        logger.info(
+            "Adding implicit conversion for input '%s' from %s to %s",
+            inpt.name,
+            to_mime(stored_format, official=False),
+            to_mime(inpt.datatype, official=False),
+        )
+        in_file = sourced.pop(inpt.name)
+        if is_union(stored_format):
+            if all(
+                is_coercible(inpt.datatype, ff) for ff in ty.get_args(stored_format)
+            ):
+                # No conversion is ever required
+                continue
+            # We need to use a workflow to determine the actual converter that needs to
+            # be used at runtime when dealing with union column datatypes
+            converter_task = RuntimeConverterWorkflow(
+                datatype=inpt.datatype,
+                converter_args=converter_args.get(inpt.name, {}),
             )
-            in_file = sourced.pop(inpt.name)
+            in_file_name = "in_file"
+            out_file_name = "out_file"
+        else:
+            # The source -> input conversion is fixed so we know which converter to use at
+            # build time
+            converter = inpt.datatype.get_converter(stored_format)
             converter_task = copy(converter.task)
+            in_file_name = converter.in_file
+            out_file_name = converter.out_file
             for nm, val in converter_args.get(inpt.name, {}).items():
                 setattr(converter_task, nm, val)
-            if ty.get_origin(source_types[inpt.name]) is StateArray:
-                # Iterate over all items in the sequence and convert them
-                # separately
-                converter_task.split(converter.in_file, **{converter.in_file: in_file})
-            else:
-                setattr(converter_task, converter.in_file, in_file)
-            # Insert converter
-            converter_outputs = workflow.add(
-                converter_task, name=f"{inpt.name}_input_converter"
-            )
-            # Map converter output to input_interface
-            sourced[inpt.name] = getattr(converter_outputs, converter.out_file)
+        # Split converter input if state array
+        if ty.get_origin(source_types[inpt.name]) is StateArray:
+            # Iterate over all items in the sequence and convert them
+            # separately
+            converter_task.split(in_file_name, **{in_file_name: in_file})
+        else:
+            setattr(converter_task, in_file_name, in_file)
+        # Add converter to workflow
+        converter_outputs = workflow.add(
+            converter_task, name=f"{inpt.name}_input_converter"
+        )
+        # Map converter output to input_interface
+        sourced[inpt.name] = getattr(converter_outputs, out_file_name)
 
     # Copy the task of the pipeline that actually performs the analysis/processing and
     # connections from the sourced/converted inputs
@@ -532,8 +551,8 @@ def PipelineRowWorkflow(
         sink_name = path2varname(outpt.name)
         if (
             outpt.datatype
-            and not issubclass(outpt.datatype, stored_format)
-            and not issubclass(stored_format, outpt.datatype)
+            and not TypeParser.is_subclass(outpt.datatype, stored_format)
+            and not TypeParser.is_subclass(stored_format, outpt.datatype)
         ):
             converter = stored_format.get_converter(outpt.datatype)
             logger.info(
@@ -636,8 +655,8 @@ def SourceItems(
     logger.debug("Sourcing %s", inputs)
     sourced: ty.Dict[str, ty.Union["frametree.core.row.DataRow", DataType]] = {}
     missing_inputs: ty.Dict[str, str] = {}
-    row = frameset.row(row_frequency, row_id)
     with frameset.store.connection:
+        row = frameset.row(row_frequency, row_id)
         for inpt in inputs:
             # If the required datatype is of type DataRow then provide the whole
             # row to the pipeline input
@@ -684,13 +703,50 @@ def SinkItems(
     if provenance is not None:
         raise NotImplementedError("Provenance storage not implemented yet")
     logger.debug("Sinking %s", items)
-    row = frameset.row(row_frequency, row_id)
     with frameset.store.connection:
+        row = frameset.row(row_frequency, row_id)
         for outpt_name, output in items.items():
             row.cell(outpt_name).item = output
     return row_id
 
 
+@workflow.define(outputs=["out_file"])
+def RuntimeConverterWorkflow(
+    in_file: FileSet,
+    datatype: ty.Type[FileSet],
+    converter_args: ty.Dict[str, ty.Any],
+) -> FileSet:
+    """A workflow that selects the appropriate converter for a union datatype
+    at runtime based on the actual type of the input file.
+
+    Parameters
+    ----------
+    in_file : DataType
+        the input file to be converted
+    datatype : type
+        the target datatype to convert to
+    converter_args : dict
+        keyword arguments passed on to the converter to control how the
+        conversion is performed.
+
+    Returns
+    -------
+    out_file : DataType
+        the converted output file
+    """
+    if is_coercible(type(in_file), datatype):
+        return in_file  # type: ignore[return-value]
+    converter = datatype.get_converter(type(in_file))
+    task = attrs.evolve(converter.task, **converter_args)
+    setattr(task, converter.in_file, in_file)
+    out = workflow.add(task)
+    return getattr(out, converter.out_file)
+
+
+def is_coercible(t: ty.Type[DataType], u: ty.Type[DataType]) -> bool:
+    return (issubclass(t, u) or issubclass(u, t)) and not (is_union(u) or is_union(t))
+
+
 # Provenance mismatch detection methods salvaged from data.provenance
 
 # def mismatches(self, other, include=None, exclude=None):
