@@ -13,8 +13,9 @@ from frametree.core.serialize import (
 )
 import frametree
 from ..axes import Axes
-from fileformats.core import DataType
+from fileformats.core import DataType, FileSetPrimitive, FieldPrimitive
 from fileformats.text import Plain as PlainText
+from pydra.utils.typing import is_union
 from frametree.core.utils import (
     get_config_file_path,
     NestedContext,
@@ -31,6 +32,7 @@ from frametree.core.exceptions import (
 S = ty.TypeVar("S", bound="Store")
 DT = ty.TypeVar("DT", bound="DataType")
 
+
 logger = logging.getLogger("frametree")
 
 
@@ -287,6 +289,8 @@ class Store(metaclass=ABCMeta):
         name : str, optional
             name of the dataset definition, which distinguishes it from alternative
             definitions on the same data, by default None
+        **kwargs:
+            Additional keyword arguments to pass to the FrameSet constructor
 
         Returns
         -------
@@ -711,7 +715,9 @@ class Store(metaclass=ABCMeta):
         """
 
     @abstractmethod
-    def get(self, entry: DataEntry, datatype: ty.Type[DT]) -> DT:
+    def get(
+        self, entry: DataEntry, datatype: ty.Type[DT]
+    ) -> FileSetPrimitive | FieldPrimitive:
         """
         Gets the data item corresponding to the given entry
 
