@@ -1,32 +1,34 @@
 from __future__ import annotations
-import os
-import os.path as op
-import typing as ty
-from pathlib import Path
-from abc import abstractmethod
-import time
-import logging
+
 import errno
 import json
+import logging
+import os
+import os.path as op
 import shutil
+import time
+import typing as ty
+from abc import abstractmethod
+from pathlib import Path
+
 import attrs
-from fileformats.core import DataType, FileSet, Field
+from fileformats.core import DataType, Field, FieldPrimitive, FileSet, FileSetPrimitive
 from fileformats.generic import File
+from pydra.utils.typing import TypeParser
+
+from frametree.core.exceptions import DatatypeUnsupportedByStoreError, FrameTreeError
 from frametree.core.utils import (
-    dir_modtime,
     JSON_ENCODING,
     append_suffix,
+    dict_diff,
+    dir_modtime,
+    full_path,
 )
-from frametree.core.exceptions import (
-    FrameTreeError,
-    DatatypeUnsupportedByStoreError,
-)
-from frametree.core.utils import dict_diff, full_path
+
 from ..entry import DataEntry
 from ..row import DataRow
 from .base import Store
 
-
 logger = logging.getLogger("frametree")
 
 
@@ -178,9 +180,7 @@ class RemoteStore(Store):
         """
 
     @abstractmethod
-    def download_value(
-        self, entry: DataEntry
-    ) -> ty.Union[float, int, str, ty.List[float], ty.List[int], ty.List[str]]:
+    def download_value(self, entry: DataEntry) -> FieldPrimitive:
         """
         Extract and return the value of the field from the store
 
@@ -198,7 +198,7 @@ class RemoteStore(Store):
     @abstractmethod
     def upload_value(
         self,
-        value: ty.Union[float, int, str, ty.List[float], ty.List[int], ty.List[str]],
+        value: FieldPrimitive,
         entry: DataEntry,
     ) -> None:
         """Store the value for a field in the XNAT repository
@@ -292,7 +292,9 @@ class RemoteStore(Store):
     # Abstractmethod implementations
     ################################
 
-    def get(self, entry: DataEntry, datatype: ty.Type[DT]) -> DT:
+    def get(
+        self, entry: DataEntry, datatype: ty.Type[DT]
+    ) -> FileSetPrimitive | FieldPrimitive:
         with self.connection:
             if entry.datatype.is_fileset:
                 item = self.get_fileset(entry, datatype)
@@ -300,10 +302,11 @@ class RemoteStore(Store):
                 item = self.get_field(entry, datatype)
             else:
                 raise DatatypeUnsupportedByStoreError(entry.datatype, self)
-        assert isinstance(item, datatype)
         return item
 
     def put(self, item: DT, entry: DataEntry) -> DT:
+        if not isinstance(item, entry.datatype):
+            item = entry.datatype(item)
         with self.connection:
             if entry.datatype.is_fileset:
                 item = self.put_fileset(item, entry)
@@ -392,7 +395,7 @@ class RemoteStore(Store):
     # Get and putters #
     ###################
 
-    def get_fileset(self, entry: DataEntry, datatype: type) -> FileSet:
+    def get_fileset(self, entry: DataEntry, datatype: type) -> list[Path]:
         """
         Caches a fileset to the local file system and returns the path to
         the cached files
@@ -458,7 +461,7 @@ class RemoteStore(Store):
                     str(cache_path) + self.CHECKSUM_SUFFIX, "w", **JSON_ENCODING
                 ) as f:
                     json.dump(checksums, f, indent=2)
-        return datatype(cache_path.iterdir())
+        return list(cache_path.iterdir())
 
     def put_fileset(self, fileset: FileSet, entry: DataEntry) -> FileSet:
         """
@@ -526,7 +529,7 @@ class RemoteStore(Store):
         )
         return cached
 
-    def get_field(self, entry: DataEntry, datatype: type) -> Field:
+    def get_field(self, entry: DataEntry, datatype: type) -> FieldPrimitive:
         """
         Retrieves a fields value
 
@@ -540,9 +543,9 @@ class RemoteStore(Store):
         value : ty.Union[float, int, str, ty.List[float], ty.List[int], ty.List[str]]
             The value of the field
         """
-        return datatype(self.download_value(entry))
+        return self.download_value(entry)
 
-    def put_field(self, field: Field, entry: DataEntry) -> None:
+    def put_field(self, field: Field | FieldPrimitive, entry: DataEntry) -> None:
         """Store the value for a field in the XNAT repository
 
         Parameters
@@ -552,7 +555,7 @@ class RemoteStore(Store):
         value : str or float or int or bool
             the value to store
         """
-        return self.upload_value(entry.datatype(field).value, entry)
+        return self.upload_value(TypeParser(entry.datatype).coerce(field).value, entry)
 
     ##############
     # Public API #
